package com.geocities.cgbusch.hieropreter;
import java.io.*;
import java.util.*;
import java.lang.*;


/**
 * Hieropreter's Engine by Chris Busch Copyright 2002 cgbusch@yahoo.com
 *
 * <br> All code Copyright 2002 by Chris Busch cgbusch@yahoo.com<p>
 *This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 <p>
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details. <p>
 *
 * @author Chris Busch (cbusch) $Author: cbusch $
 * @since Mon Jan 07 15:27:05 2002
 * @version $Revision: 1.7 $ $Date: 2002/10/03 22:39:13 $
 * @see 
 */

public class   CodeMatch {

    public static final String TITLE
	="Hieropreter v4.01 by Chris Busch Copyright 2002 cgbusch@yahoo.com";

    protected WidthContainer aedict=new WidthContainer();
    protected WidthContainer engdict=new WidthContainer();

    /** Maps mdc sequences to gardiner symbols */
    protected WidthContainer mdc2gdict=new WidthContainer();

    /** Maps gardiner codes to list of mdc (loaded)*/
    protected Map gtom = new TreeMap();
    
    /** maps gardiner codes to Chars (auto generated)*/
    protected Map gardiners = new HashMap();

   
    public static class Flags {
	boolean showai=false, mdc2gard=false, ae2eng=true;
	Comparator logic=null;
	
	int maxLines=10;
	
	public Flags(Comparator l,boolean ae2Eng,boolean showAI) {
	    logic=l;
	    this.showai=showAI;
	    this.ae2eng=ae2Eng;
	    mdc2gard = ! ae2eng;
	}

	
	public Flags(char l,boolean ae2Eng,boolean showAI) {
	    this.showai=showAI;
	    this.ae2eng=ae2Eng;
	    mdc2gard = ! ae2eng;
	    switch(l) {
	    case 'f': this.logic = DepthList.firstFit;
		break;
	    case 'b': this.logic = DepthList.bigFit;
		break;
	    case 't': this.logic = DepthList.topFit;
		break;
	    case 's': this.logic = DepthList.straightFit;
		break;
	    default: throw new IllegalArgumentException("improper comparator");
	    }
	}
	
	public void setMaxGuesses(int m) { maxLines=m;	}

	public boolean getShowAiValues() { return showai; }



	
    }


    static  protected final char RED=(char)290;
    static  protected final char BLACK=(char)291;
    static  protected final char SHADE=(char)292;
    static  protected final char NOSHADE=(char)293;
    
    /** Acts as the start of the pseudo Unicode Hieroglyphs */
    static  final char OFFSET=(char)300;

    /** maps gardiner chars to codes (auto generated)*/
    protected List gardinerchars = new ArrayList();

    /** Constructs an empty engine */
    public CodeMatch() { 

    }

    /** Loads dict and g map from class location */
    public CodeMatch(boolean preload) throws IOException{ 
	if( preload ) {
	    loadGMap(new BufferedReader(new InputStreamReader(this.getClass().getResourceAsStream("gardiner2mdc.txt"))));
	    loadDict(new BufferedReader(new InputStreamReader(this.getClass().getResourceAsStream("egyptdictionary.txt"))));
	}
    }

    public boolean isDeterminate(String gardcode){
	if(gardcode.length()==1) return false; //its a letter?
	gardcode = gardcode.toUpperCase().trim();
	if( gardcode.equals("N(35)") ) gardcode="N35A";
	return !gtom.containsKey(gardcode);
    }

    protected boolean isDeterminate(char gardchar){
	if(gardchar<OFFSET) return false; //its a letter?
	return isDeterminate(getNormChar(gardchar));
    }

    /** return pseudo unicode char */
    protected char getGardChar(String gardcode) {
	gardcode = gardcode.toUpperCase().trim();
	if( gardcode.equals("N(35)") ) gardcode="N35A";
	//convert to J11
	if( gardcode.startsWith("AA") ) gardcode=replace(gardcode,"AA","J");
	Character c=(Character)gardiners.get(gardcode);
	if( c==null ) {
	    c=new Character((char)(OFFSET+gardinerchars.size()));
	    gardinerchars.add(gardcode);
	    gardiners.put(gardcode,c);
	}
	return c.charValue();
    }

    /** converts pseudo unicode char to gardiner code letter+number. */
    public String getNormChar(char gardchar) {
	if(gardchar<OFFSET) return ""+gardchar;
	return (String)gardinerchars.get(gardchar-OFFSET);
    }

    /** converts a gardiner code sequence to a  pseudo unicode string */
    public String getGardString(String normal) {
	StringBuffer sb=new StringBuffer();
	normal=replace(normal,"N(35)","N35a");
	normal=replace(normal,"Y1v","Y1");
	StringTokenizer st=new StringTokenizer(normal,"\t\r\n !@#$%^&*(){}[]-=_;,:?\\",true);
	boolean skipnextnum=false;
	boolean skipnext=false;
	String prev="";
	String cur="";
	while(st.hasMoreTokens()){
	    String s=cur=(String)st.nextToken();
	    if( s.equals("<") 
		|| s.equals(">") 
		|| s.equals("<0") 
		|| s.equals("<1") 
		|| s.equals("<2") 
		|| s.equals("<h1") 
		|| s.equals("<h2") 
		|| s.equals("0>") 
		|| s.equals("1>") 
		|| s.equals("2>") 
		|| s.equals("h1>") 
		|| s.equals("h2>") 
		) {
		continue;
	    }
	    if(prev.equals("#") ) {
		if( s.equals("b") || s.equals("e") || allDigits(s) ) {
		    prev="";
		    continue;
		}
	    }else if(prev.equals("$") ) {
		if( s.equals("r")) {
		    s=""+RED;
		}else if(s.equals("b")) {
		    s=""+BLACK;
		}else{
		    prev="";
		    continue;
		}
	    }else if(prev.equals("\\") ) {
		if( s.startsWith("t") || s.startsWith("r") || allDigits(s) ) {
		    s="";
		}
	    }
	    if(skipnext) { skipnext=false; continue; }
	    if(skipnextnum && hasDigit(s) ) { skipnextnum=false; continue; }
	    if( s.equals("//") || s.equals("/") || s.equals("h/") || s.equals("v/")) {
		s="?";
	    }else if(s.length()>1 && Character.isLetter(s.charAt(0)) && hasDigit(s) ) {
		//gardiner sign
		sb.append(getGardChar(s));
	    }else if(s.length()>1){
		sb.append(s);
	    }else if(s.length()==1){
		switch( s.charAt(0) ) {
		    ///////discard these chars
		case '(':
		case ')':
		case ':':
		case '*':
		case '#':
		case '&':
		case '\\':
		case '$':
		    s="";
		    break;
		default:
		    //keep char
		}
		sb.append(s);
	    }
	    prev=cur;
	}
	while(sb.length()>0 && sb.charAt(0) == '-') { 
	    sb.deleteCharAt(0); 
	}
	while(sb.length()>0 && sb.charAt(sb.length()-1) == '-') { 
	    sb.setLength(sb.length()-1); 
	}
	String res=sb.toString();
	res=replace(res,"--","-"); //lame
	res=replace(res,"--","-"); //lame
	res=replace(res,"--","-"); //lame
	return res;
    }

    protected boolean hasDigit(String in) {
	for(int i=0;i<in.length();i++){
	    if( Character.isDigit(in.charAt(i)) ) return true;
	}
	return false;
    }

    protected boolean allDigits(String in) {
	for(int i=0;i<in.length();i++){
	    if( !Character.isDigit(in.charAt(i)) ) return false;
	}
	return true;
    }

    /** converts to a gardiner code sequence from a  pseudo unicode string */
    protected String getNormalString(String gardchars) {
	StringBuffer sb=new StringBuffer();
	for(int i=0;i<gardchars.length();i++){
	    sb.append(getNormChar(gardchars.charAt(i)));
	}
	return sb.toString();
    }

    public void addEnglish(String eng,Word w){
	StringTokenizer st=new StringTokenizer(eng,"\t\r\n !@#$%^&*(){}[]-=_;,:?\\",false);
	while(st.hasMoreTokens()){
	    String s=(String)st.nextToken();
	    engdict.add(s,w);
	}
    }

    /** returns a IMG tag */
    public String getGardImages(String gstr) {
	StringBuffer newvalue=new StringBuffer();
	for(int i=0;i<gstr.length();i++) {
	    char c=gstr.charAt(i);
	    if( c>=OFFSET ) {
		String gs2 = getNormChar(c); 
		newvalue.append("<img src=\"images/"+gs2.toLowerCase()
				+".gif\" alt=\"\"><font size=-2>"+gs2+"</font> ");
	    }else{
		newvalue.append(c);
	    }
	}
	return newvalue.toString();
    }


    public void loadDict(BufferedReader in) throws java.io.IOException {
        String input;
	int linenum=0;
        while(((input=in.readLine())!=null )) {
	    linenum++;
	    String line=input.trim();
	    List rec=split("%",line);
	    if( rec.size()!=6 ) throw new IOException("Improper format "+linenum+" "+line);
	    String mdc=(String)rec.remove(0);
	    String gards=(String)rec.remove(0);
	    String type=(String)rec.remove(0);
	    String eng=(String)rec.remove(0);
	    String src=(String)rec.remove(0);
	    int prob=Integer.parseInt(((String)rec.remove(0)).trim());
	    String gstr=getGardString(replace(gards.trim()," ",":"));
	    String gardimg="<BR>"+getGardImages(gstr);
	    mdc=mdc.trim();
	    eng=eng.trim();
	    //eng += " L"+linenum;
	    //store english
	    addEnglish(eng.toLowerCase(),new Word(mdc,gardimg,type,eng,src,0));
	    //store by mdc 
	    if( mdc.length()>0 ) {
		aedict.add(mdc,new Word(mdc,gardimg,type,eng,src,
					prob+800*mdc.length()+gstr.length()));
	    }
	    //convert gardiner logic
	    if( gstr.length() > 0 ) {
		//System.out.println("gs:"+getNormalString(gstr));
		aedict.add(gstr,new Word(mdc,gardimg,type,eng,src,
					 prob+800*gstr.length()+400*mdc.length()));
		//load gard with -s
		{
		    String gstr2=getGardString(replace(gards.trim()," ","-"));
		    if( !gstr2.equals(gstr)) {
			aedict.add(gstr2,new Word(mdc,gardimg,type,eng,src,
						  prob+800*gstr.length()+400*mdc.length()));
		    }
		}
		///////load gardiner as mdc
		String[] r=convertGardToMdc(gstr);
		if( !gstr.equals(r[0]) ) {
		    aedict.add(r[0],new Word(mdc,gardimg,type,eng,
					     src,prob+400*(gstr.length()+mdc.length())));
		}
		if( ! gstr.equals(r[1]) && ! r[1].equals(r[0]) ) {
		    aedict.add(r[1],new Word(mdc,gardimg,type,eng,
					     src,prob+400*(gstr.length()+mdc.length())));
		}
		//load without determinants
		int lastdet=0;
		for(int stop=gstr.length()-1;stop>=0;stop--) {
		    if(!isDeterminate(gstr.charAt(stop))){
			lastdet=stop;
			break;
		    }
		}
		if( lastdet>0 && lastdet<gstr.length()-1) {
		    //trimmed determinates
		    int diff = gstr.length()-lastdet+1;
		    String gstrtrim=gstr.substring(0,lastdet+1);
		    aedict.add(gstrtrim,new Word(mdc,gardimg,type,eng,src,
						 prob+(200-diff)*gstrtrim.length()));
		}
	    }
	    //make something easy to look up
	    String mstr=replace(mdc,"jj","y");
	    mstr=replace(mstr,"j","i");
	    mstr=replace(mstr,"z","s");
	    mstr=discard(mstr,"-/.()^=");
	    if( !mstr.equals(mdc) && mstr.length()>0 ) {
		int diff = mdc.length()-mstr.length();
		aedict.add(mstr,new Word(mdc,gardimg,type,eng,src,prob+(500-diff)*mstr.length()+gstr.length()));
	    }
        }
        in.close();
    }

    public void loadGMap(BufferedReader in) throws java.io.IOException {
        String input;
	int linenum=0;
        while(((input=in.readLine())!=null )) {
	    linenum++;
	    String line=input.trim();
	    if(line.startsWith("#")) continue;
	    List rec=split(",",replace(line.trim()," ",""));
	    String gard=((String)rec.remove(0)).trim().toUpperCase();
	    gtom.put(gard,rec);
	    //populate the dictionary to convert mdc sequences to gardiner sc
	    String gstr=getGardString(gard);
	    String gardimg=getGardImages(gstr);
	    String eng ="Gardiner sign "+gard+" with MdC of "+rec;
	    mdc2gdict.add(gstr,new Word("",gardimg,"symbol",eng,"Gardiner",2));
	    Iterator letters = rec.iterator();
	    while(letters.hasNext()) {
		String mdc=letters.next().toString();
		mdc2gdict.add(mdc,new Word(mdc,gardimg,"symbol",eng,"Gardiner",1));
	    }
        }
        in.close();
	//System.out.println(gardinerchars);
    }


    static protected String replace(String value,String match,String sub) {
	StringBuffer newvalue=new StringBuffer();
	int in=value.indexOf(match);
	if( in<0 ) return value; //dont make unneeded strings
	while( (in= value.indexOf(match))>=0 ) {
	    newvalue.append(value.substring(0,in));
	    newvalue.append(sub);
	    value=value.substring(in+match.length());
	}
	newvalue.append(value);
	return newvalue.toString();
    }

    static protected String discard(String in,String matchany) {
	StringBuffer newvalue=new StringBuffer();
	for(int i=0;i<in.length();i++) {
	    if( matchany.indexOf(in.charAt(i)) >=0 ) continue;
	    newvalue.append(in.charAt(i));
	}
	return newvalue.toString();
    }
    

    public List findByEnglish(String eng) {
	System.out.println("english input "+(new java.util.Date()) +":"+eng); 
	return engdict.lookup(eng.trim().toLowerCase());
    }

    /** The first string returned is the maximal and the 2nd is the minimal "phoneme" char set.*/
    protected String[] convertGardToMdc(String test) {
	String newstr="";
	String newstr2="";
	String p=" ";
	for(int ii=test.length()-1; ii>=0; ii--) {
	    char c = test.charAt(ii);
	    if( c>=OFFSET ) {
		String found=" "; //mmmm
		List lit=(List)gtom.get(getNormChar(c));
		if( lit == null || lit.size()==0 ) {
		    //preserve the symbol
		    found = ""+c;
		}else if( p.length()==1 ) {
		    for(int f=lit.size()-1;f>=0;f--) {
			found=((String)lit.get(f)).trim();
			if( found.endsWith(p) ) {
			    break;
			}
		    }
		}else if( lit.size()>=1 ) {
		    //take the first  one :(
		    found=((String)lit.get(0)).trim();
		}
		//need to add chopping of newstr if found endsWith(p) if p.len==1
		if( found.length()>1 && p.length()==1 && found.endsWith(p) ) {
		    if( newstr2.startsWith("-") ) newstr2=newstr2.substring(1);
		    newstr2=found.substring(0,found.length()-1)+newstr2; //ouch
		}else{
		    if( "".equals(newstr2) ) {
			newstr2=found;
		    }else{
			newstr2=found+( newstr.startsWith("-") ? "" : "-" )+newstr2; //ouch
		    }
		}
		if( "".equals(newstr) ) {
		    newstr=found;
		}else{
		    newstr=found+( newstr.startsWith("-") ? "" : "-")+newstr; //ouch
		}
		p=found;
	    }else{
		newstr=c+newstr; //ouch
		newstr2=c+newstr2; //ouch
		if( Character.isLetter(c) ) {
		    p=""+c;
		}
	    }
	}	
	return new String[] {newstr,newstr2};
    }

    /** Takes gardiner codes and expands them to mdc */
    protected void findExpandedWord(String test,DepthList dl,int start,int len) {
	////////////////////////
	//let's translate to mdc
	String[] r=convertGardToMdc(test);
	String newstr=r[0];
	String newstr2=r[1];
	//System.out.println("test:"+getNormalString(test)+" n1:"+getNormalString(newstr)+" n2:"+getNormalString(newstr2)	   );
	if(! test.equals(newstr) ) {
	    findRegularWord(newstr,dl,start,len,80);
	    findTrimmedWord(newstr,dl,start,len,60);
	}
	if( ! newstr.equals(newstr2) && ! test.equals(newstr2) ) {
	    findRegularWord(newstr2,dl,start,len,79);
	    findTrimmedWord(newstr2,dl,start,len,55);
	}
    }

    protected void findRegularWord(String test,DepthList dl,int start,int len,int p) {
	//System.out.println("'"+test+"'");
	Iterator fi = aedict.lookup(test).iterator();
	while(fi.hasNext()) dl.add(start,len,p,(Comparable)fi.next());
    }


    /** Takes gardiner codes and expands them to mdc */
    protected void findTrimmedWord(String test,DepthList dl,int start,int len,int p) {
	////////////////////////
	//let's translate to mdc
	String newstr=discard(test,"-/.()^");
	if( newstr.equals(test)==false && newstr.length()>0) {
	    findRegularWord(newstr,dl,start,len,p);
	}
	String newstr2 = replace(newstr,"jj","y");
	newstr2 = replace(newstr2,"j","i");
	newstr2 = replace(newstr2,"z","s");
	if( newstr2.equals(newstr)==false && newstr2.length()>0) {
	    findRegularWord(newstr2,dl,start,len,p-10);
	}
    }

    /** Searches mdc2gdict dictionary finding the gardiner symbols that match the mdc */
    public DepthList findMdcSymbols(String input,DepthList dl) {
	int max=mdc2gdict.getMaxLength()*4; //*4 due to char elimination
	if( max>input.length()) max=input.length();
	for(int len=max;len>=1;len--) {
	    for(int i=0;i<input.length()-len+1;i++) {
		String test=input.substring(i,i+len);
		if( test.startsWith("-") ) continue;
		if( test.endsWith("-") ) continue;
		if( test.indexOf('!')>=0 ) continue;
		if( test.indexOf(RED)>=0 ) continue;
		if( test.indexOf(BLACK)>=0 ) continue;
		Iterator fi = mdc2gdict.lookup(test).iterator();
		while(fi.hasNext()) dl.add(i,len,1,(Comparable)fi.next());
	    }
	}
	return dl;
    }

    /** Attempts to fill the DepthList with the ENGLISH Word definitions given AE. */
    public DepthList findWords(String input,DepthList dl) {
	int max=aedict.getMaxLength()*4; //*4 due to char elimination
	if( max>input.length()) max=input.length();
	for(int len=max;len>=1;len--) {
	    for(int i=0;i<input.length()-len+1;i++) {
		String test=input.substring(i,i+len);
		//System.out.println("'"+getNormalString(test)+"'");
		if( test.startsWith("-") ) continue;
		if( test.endsWith("-") ) continue;
		if( test.indexOf('!')>=0 ) continue;
		if( test.indexOf(RED)>=0 ) continue;
		if( test.indexOf(BLACK)>=0 ) continue;
		//if( test.indexOf(' ')>=0 ) continue;
		findRegularWord(test,dl,i,len,100);
		findTrimmedWord(test,dl,i,len,75);
		findExpandedWord(test,dl,i,len);
		//System.out.println("dl="+dl);
	    }
	}
	return dl;
    }



    /** Translates a paragraph of AE. High level method that starts the ball rolling. */
    public String translate(String input,Flags flags) {
	System.out.println("translate input "+(new java.util.Date()) +":"+input);
	if( input.length() > 1000 ) return "<p> Input text is too long, you are only allowed to submit 1000 characters at a time.";
	//StringWriter out=new StringWriter();
	StringBuffer out=new StringBuffer();
	StringTokenizer st=new StringTokenizer(input,"!");
	while(st.hasMoreTokens()){
	    String in=st.nextToken();
	    if(in.length()>200) {
		out.append("<p> Input text is too long, please split it up by ! marks.");
	    }else{
		showWords(out,in,flags);
	    }
	}
	return 	TITLE+"<p>"+out.toString();
    }

    /** Translates a sentence of AE. */
    synchronized protected void showWords(StringBuffer out,String input,Flags flags) {
	String gstr=getGardString(input);
	if( flags.ae2eng ) {
	    DepthList dl=new DepthList();
	    //a little lame should delete Flags object and just use DepthList
	    if(flags.logic!=null) dl.setComparator(flags.logic);
	    dl=findWords(gstr,dl);
	    out.append("<p><b>Input sequence with up to "+flags.maxLines+" rows of English guesses.</b><br>\n");
	    display(out,gstr,dl,flags);
	}
	//////////////convert mdc to gardiner codes
	if( flags.mdc2gard ) {
	    DepthList dlsymbols=new DepthList();
	    dlsymbols=findMdcSymbols(gstr,dlsymbols);
	    out.append("<p><b>MdC sequence converted to Gardiner Sign Codes</b><br>\n");
	    display(out,gstr,dlsymbols,flags);
	}
    }

    protected static List split(String delim,String s) {
	List v = new ArrayList();
	int i = 0;
	while((i=s.indexOf(delim)) > -1) {
	    v.add(s.substring(0,i));
	    s = s.substring(i+delim.length());
	}
	v.add(s);
	return v;
    }

    protected String toString(String in) {
	StringBuffer newvalue=new StringBuffer();
	for(int i=0;i<in.length();i++) {
	    char c=in.charAt(i);
	    if( c>=OFFSET ) {
		newvalue.append("["+(int)c+"]");
	    }else{
		newvalue.append(c);
	    }
	}
	return newvalue.toString();
    }
    
    private char curcolor=BLACK;
    private char curshade=NOSHADE;

    protected String getCell(char i) {
	String s=" ";
	if(i==RED) curcolor=RED;
	else if(i==BLACK) curcolor=BLACK;
	else if(i==SHADE) curshade=SHADE;
	else if(i==NOSHADE) curshade=NOSHADE;
	else{
	    String norm = getNormChar(i);
	    if( hasDigit(norm) ) {
		s="<img src=\"images/"+norm.toLowerCase()+".gif\" alt=\"\">"+norm;
		List mdc=(List)gtom.get(norm);
		if( mdc!=null ) {  s+=" <i>"+mdc+"</i>"; } 
	    }else{
		s=norm;
	    }
	}
	if( curcolor==RED ) {
	   s="<font color=red>"+s+"</font>";
	}
	//String cbox="<input typ
	return "<td"+(curshade==SHADE ? " bgcolor=EEEEEE " : "" )+">"+s+"</td>\n";
    }

    /** generates html */
    protected void display(StringBuffer o,String in,DepthList dl,Flags flags) {
	o.append("<table border=1>");
	for(int i=0; i<=dl.getMaxDepth() && i<flags.maxLines;i++) {
	    ///orig
	    if( i % 5 == 0 ) {
		o.append("<tr bgcolor=FFFFDD><td>Original Text:</td>\n");
		for(int ii=0;ii<in.length();ii++) {
		    o.append(getCell(in.charAt(ii)));
		}
		o.append("</tr>");
	    }
	    ///guesses
	    o.append("<tr valign=top bgcolor=white>");
	    if( i==0 ) {
		o.append("<td><i>Best Guess:</i></td>");
	    }else if( i==1 ) {
		o.append("<td><i>Second Guess</i></td>");
	    }else if( i==2 ) {
		o.append("<td><i>Third Guess</i></td>");
	    }else{
		o.append("<td><i>"+(i+1)+"th Guess</i></td>");
	    }
	    List d=dl.getDepth(i);
	    Iterator di=d.iterator();
	    int pos=0;
	    while( di.hasNext()) {
		DepthListNode n=(DepthListNode)di.next();
		for(int j=pos;j<n.getStart();j++) { 	
		    o.append("<td>");
		    // o.append("j="+j+" i="+i+""); 
		    o.append("</td>\n");
		}
		o.append("\n<td colspan=\""+n.length+"\">\n");
		o.append(n.word);
		if( flags.getShowAiValues() ) {
		    o.append(" <font color=EEEEEE>Q="+((Word)n.word).getQuality()+" S="+n.getStart()+" L="+n.length+" D="+i+" P="+n.priority+" sc="+n.score+"</font>");
		}
		o.append("</td>\n");
		pos = n.start+n.length;
	    }
	    o.append("</tr>\n");
	}
	o.append("</table><p>\n");
    }

    /** generates html */
    protected void displayVert(PrintWriter o,String in,DepthList dl) {
	//displayHeader(o,in,dl,2);
	o.println("<p>Analysis: (Sentence is vertical along left side)<br><table border=1>");
	for(int i=0;i<in.length();i++) {
	    o.println("<tr>");
	    o.println(getCell(in.charAt(i)));
	    List d=dl.getStart(i);
	    Iterator di=d.iterator();
	    int pos=0;
	    while( di.hasNext()) {
		DepthListNode n=(DepthListNode)di.next();
		//for(int j=pos;j<n.depth;j++) { 	o.println("<td></td>"); }
		o.println("<td rowspan=\""+n.length+"\">");
		o.print(n.word); //+" d="+n.depth+" s="+n.start+" l="+n.length);
		o.println("</td>");
		pos = n.depth+1;
	    }
	    o.println("</tr>");
	}
	o.println("</table>");
	o.flush();
    }
    


    protected void displayText(String in,DepthList dl) {
        for(int i=0;i<=dl.getMaxDepth();i++) {
            List d=dl.getDepth(i);
            Iterator di=d.iterator();
            while( di.hasNext()) {
                DepthListNode n=(DepthListNode)di.next();
                for(int j=0;j<n.start;j++) {                System.out.print(' '); }
                System.out.println(n.word+":"+n.depth);
            }
        }
    }   


    /**
     * Command line version, a little slow since the files need to be loaded each time its ran...
     *
     */
    public static void main(String[] args) throws Exception {
	System.err.println("usage args: aedict-file gardiner-map-file (f|b|t|s) t 'AE words to translate'  > x.html");
	System.err.println("usage args: aedict-file gardiner-map-file (f|b|t|s) c 'convert MdC to glyphs'  > x.html");
	System.err.println("usage args: aedict-file gardiner-map-file (f|b|t|s) e 'english word' > y.html");
        //put your test code here
        CodeMatch cm=new CodeMatch();
	cm.loadGMap(new BufferedReader(new FileReader(args[1])));
	cm.loadDict(new BufferedReader(new FileReader(args[0])));
	if( "e".equals(args[3] ) ) {
	    Iterator i=cm.findByEnglish(args[4]).iterator();
	    while(i.hasNext()) {
		System.out.println(i.next());
	    }
	}else {
	    CodeMatch.Flags flags=new CodeMatch.Flags(args[2].charAt(0),
						      "t".equals(args[3]),
						      true
						      );  
	    System.out.println(cm.translate(args[4],flags ));
	}
    }
  
}
