package com.geocities.cgbusch.hieropreter;
import java.io.*;
import java.util.*;

/**
 * Data structure to represent the 2 dimensional results.
 *
 * <br> All code Copyright 2002 by Chris Busch cgbusch@yahoo.com
 * @author Chris Busch (cbusch) $Author: cbusch $
 * @since Mon Jan 07 15:27:05 2002
 * @version $Revision: 1.3 $ $Date: 2002/09/19 23:04:29 $
 * @see 
 */

public class   DepthList extends ArrayList {

    private int maxdepth=-1, maxstart=-1;
    private boolean needsort=true;

    public final static Comparator depthComparator= new DepthComparator();
    public final static Comparator startComparator = new StartComparator();

    /** This Comparator is the essential component for the artificial intelligence decision making. */
    public final static Comparator firstFit = new FirstFit();

    /** This Comparator is the essential component for the artificial intelligence decision making. */
    public final static Comparator bigFit = new BigFit();

    /** This Comparator is the essential component for the artificial intelligence decision making. */
    public final static Comparator topFit = new TopFit();

    /** This Comparator is the essential component for the artificial intelligence decision making. */
    public final static Comparator straightFit = new StraightFit();

    /** first fit greedy algo */
    public static class FirstFit implements Comparator {
	void score(DepthListNode a) {
	    if( a.score==-1) {
		a.score=a.length*a.priority*a.getQuality()-a.start;
	    }
	}
	public int compare(Object ao,Object bo) {
	    DepthListNode a=(DepthListNode)ao;
	    DepthListNode b=(DepthListNode)bo;
	    score(a);
	    score(b);
	    int i = a.start - b.start;
	    if( i==0 ) return b.score - a.score;
	    return i;
	}
	public String toString() { return "First Fit"; }
    }

    /** largest word fit greedy algo */
    public static class BigFit implements Comparator {

	public String toString() { return "Big Fit"; }

	void score(DepthListNode a) {
	    if( a.score==-1) {
		a.score=a.length*a.priority*a.getQuality()-a.start;
	    }
	}
	public int compare(Object ao,Object bo) {
	    DepthListNode a=(DepthListNode)ao;
	    DepthListNode b=(DepthListNode)bo;
	    score(a);
	    score(b);
	    int i = b.length - a.length;
	    if( i==0 ) return b.score - a.score;
	    return i;
	}
    }

    /**  fit by biggest score with length algo */
    public static class TopFit implements Comparator {

	public String toString() { return "Top Fit"; }

	void score(DepthListNode a) {
	    if( a.score==-1) {
		a.score=a.length*100*a.priority*a.getQuality()-a.start;
	    }
	}
	public int compare(Object ao,Object bo) {
	    DepthListNode a=(DepthListNode)ao;
	    DepthListNode b=(DepthListNode)bo;
	    score(a);
	    score(b);
	    return b.score - a.score;
	}
    }

    /** fit by score */
    public static class StraightFit implements Comparator {
	
	public String toString() { return "Straight Fit"; }

	void score(DepthListNode a) {
	    if( a.score==-1) {
		a.score=a.length+a.priority*a.getQuality()-a.start;
	    }
	}
	public int compare(Object ao,Object bo) {
	    DepthListNode a=(DepthListNode)ao;
	    DepthListNode b=(DepthListNode)bo;
	    score(a);
	    score(b);
	    return b.score - a.score;
	}
    }

    private Comparator comp=null;

    private List depths=new ArrayList();

    public void setComparator(Comparator c) {
	comp=c;
    }

    /**
     * 
     */
    public DepthList() { //constructor
	super();
    }

    /** Returns the maximum depth which getDepth will work, returns -1 if empty. */
    public int getMaxDepth() { 
	if(needsort) sort();
	return maxdepth; 
    }

    /** Returns the maximum start which getStart will work, returns -1 if empty. */
    public int getMaxStart() { 
	if(needsort) sort();
	return maxstart; 
    }

    /** List of Nodes */
    public List getDepth(int d) {
	if(needsort) sort();
	List l=(List)depths.get(d);
	Collections.sort(l,startComparator);
	//System.out.println("d="+d+" ret="+l);
	return l;
    }

    /** List of Nodes */
    public List getStart(int d) {
	if(needsort) sort();
	List l=new ArrayList();
	Iterator i=iterator();
	while(i.hasNext()) {
	    DepthListNode n=(DepthListNode)i.next();
	    if( n.start==d) {
		l.add(n);
	    }
	}
	Collections.sort(l,depthComparator);
	return l;
    }

    /** only add if unique */
    public boolean add(int s,int len,int p,Comparable w) {
	DepthListNode newNode=new DepthListNode(s,len,p,w);
	this.add(newNode);
	needsort=true;
	return true;
    }


    static protected boolean miss(int astart,int aend,int bstart,int bend) {
	return ( aend<bstart || bend<astart );
    }

    public void sort() {
	depths.clear();
	maxdepth=-1;
	maxstart=-1;
	needsort=false;
	if(comp==null)	Collections.sort(this);
	else Collections.sort(this,comp);
	//System.out.println(this);
	Iterator x=this.iterator();
    FORGETIT:
	while(x.hasNext()){
	    DepthListNode newNode=(DepthListNode)x.next();
	    int start=newNode.start;
	    int len=newNode.length;
	    int end=start+len-1;
	    int depth=-1;
	    List adepth=null;
	DEPTHFIND:
	    for(int d=0;d<depths.size();d++) {
		adepth=(List)depths.get(d);
		///////check to see if this is a duplicate entry found
		if( adepth.contains(newNode) ) continue FORGETIT;
		Iterator i=adepth.iterator();
		while(i.hasNext()) {
		    DepthListNode n=(DepthListNode)i.next();
		    if( ! miss(start,end,n.start,n.start+n.length-1) ) {
			continue DEPTHFIND;
		    }
		}
		depth=d;
		break;
	    }
	    if( depth == -1) {
		adepth=new ArrayList();
		depth=depths.size();
		depths.add(adepth);
	    }
	    newNode.setDepth(depth);
	    adepth.add(newNode);
	    if( depth>maxdepth) maxdepth=depth;
	    if( start>maxstart) maxstart=start;
	}
	//System.out.println("maxs="+maxstart+" md="+maxdepth+" ds="+depths);
    }
}
